using System;
using System.Text;
using System.Drawing;
using System.IO.Ports;
using System.Windows.Forms;

using terminal;

//*****************************************************************************************
//                           LICENSE INFORMATION
//*****************************************************************************************
//   PCCom.SerialCommunication Version 1.0.0.0
//   Class file for managing serial port communication
//
//   Copyright (C) 2007  
//    Richard L. McCutchen
//   Email: richard@psychocoder.net
//   Created: 20OCT07
//
//   This program is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program.  If not, see <http://www.gnu.org/licenses/>.
//*****************************************************************************************
namespace PCComm
{
    class CommunicationManager
    {
        #region Manager Enums
        /// <summary>
        /// enumeration to hold our transmission types
        /// </summary>
        public enum TransmissionType { Text, Hex, HT }

        /// <summary>
        /// enumeration to hold our message types
        /// </summary>
        public enum MessageType { Incoming, Outgoing, Normal, Warning, Error };
        #endregion

        #region Manager Variables
                
        //property variables
        private string _baudRate = string.Empty;
        private string _parity = string.Empty;
        private string _stopBits = string.Empty;
        private string _dataBits = string.Empty;
        private string _portName = string.Empty;
        // sets the number of bytes needed in the internal buffer before a received data msg is fired
        private int _receivedBytes = 1;
        
        private TransmissionType _transType;
        private RichTextBox _displayWindow;

        private TextBox _actualPosition;
        private TextBox _leftLimit;
        private TextBox _rightLimit;

        //global manager variables
        private Color[] MessageColor = { Color.Blue, Color.Green, Color.Black, Color.Orange, Color.Red };
        private SerialPort comPort = new SerialPort();
        
        private byte[] comBuffer = new byte[100];
       
/*      private static int comBufferStart;
        private static int comBufferEnd;
*/      
        private byte[] answer = new byte[FrameLength];
        private int byte_to_int;

        static readonly byte[] Version = {1, 72, 84, 32, 86, 48, 46, 48, 54};
        static readonly byte[] State   = {1, 2, 100, 5, 0, 0, 0, 1, 109};

        // sets the frame length of the protocol
        const int FrameLength = 9;
        
        #endregion

        #region GetVersion
        public string GetVersion()
        {
            System.Text.ASCIIEncoding enc = new System.Text.ASCIIEncoding();
            return enc.GetString(Version, 4, 5);            
        }
        #endregion

        #region Manager Properties
        /// <summary>
        /// Property to hold the BaudRate
        /// of our manager class
        /// </summary>
        public string BaudRate
        {
            get { return _baudRate; }
            set { _baudRate = value; }
        }

        /// <summary>
        /// property to hold the Parity
        /// of our manager class
        /// </summary>
        public string Parity
        {
            get { return _parity; }
            set { _parity = value; }
        }

        /// <summary>
        /// property to hold the StopBits
        /// of our manager class
        /// </summary>
        public string StopBits
        {
            get { return _stopBits; }
            set { _stopBits = value; }
        }

        /// <summary>
        /// property to hold the DataBits
        /// of our manager class
        /// </summary>
        public string DataBits
        {
            get { return _dataBits; }
            set { _dataBits = value; }
        }

        /// <summary>
        /// property to hold the PortName
        /// of our manager class
        /// </summary>
        public string PortName
        {
            get { return _portName; }
            set { _portName = value; }
        }

        /// <summary>
        /// property to hold the ReceivedBytes
        /// of our manager class
        /// </summary>
        public int ReceivedBytes
        {
            get { return _receivedBytes; }
            set { _receivedBytes = value; }
        }

        /// <summary>
        /// property to hold our TransmissionType
        /// of our manager class
        /// </summary>
        public TransmissionType CurrentTransmissionType
        {
            get { return _transType; }
            set { _transType = value; }
        }

        /// <summary>
        /// property to hold our display window
        /// value
        /// </summary>
        public RichTextBox DisplayWindow
        {
            get { return _displayWindow; }
            set { _displayWindow = value; }
        }

        /// <summary>
        /// Property to hold the 
        /// of our manager class
        /// </summary>
        public TextBox ActualPosition
        {
            get { return _actualPosition; }
            set { _actualPosition = value; }
        }

        public TextBox LeftLimit
        {
            get { return _leftLimit; }
            set { _leftLimit = value; }
        }

        public TextBox RightLimit
        {
            get { return _rightLimit; }
            set { _rightLimit = value; }
        }
        #endregion

        #region Manager Constructors
        /// <summary>
        /// Constructor to set the properties of our Manager Class
        /// </summary>
        /// <param name="baud">Desired BaudRate</param>
        /// <param name="par">Desired Parity</param>
        /// <param name="sBits">Desired StopBits</param>
        /// <param name="dBits">Desired DataBits</param>
        /// <param name="name">Desired PortName</param>
        public CommunicationManager(string baud, string par, string sBits, string dBits, 
            string name, int rec, RichTextBox rtb, TextBox pos)
        {
            _baudRate = baud;
            _parity = par;
            _stopBits = sBits;
            _dataBits = dBits;
            _portName = name;
            _displayWindow = rtb;
            _actualPosition = pos;
            _receivedBytes = rec;
            //now add an event handler
            comPort.DataReceived += new SerialDataReceivedEventHandler(comPort_DataReceived);            
        }

        /// <summary>
        /// Comstructor to set the properties of our
        /// serial port communicator to nothing
        /// </summary>
        public CommunicationManager()
        {
            _baudRate = string.Empty;
            _parity = string.Empty;
            _stopBits = string.Empty;
            _dataBits = string.Empty;
            _portName = "COM1";
            _displayWindow = null;
            _actualPosition = null;
            _receivedBytes = 1;
            //add event handler
            comPort.DataReceived += new SerialDataReceivedEventHandler(comPort_DataReceived);
        }        
        #endregion

        #region WriteData
        public void WriteData(string msg)
        {
            switch (CurrentTransmissionType)
            {
                // only for Haptic Terminal use
                case TransmissionType.HT:
                    try
                    {
                        //convert the message to byte array
                        byte[] newMsg = HexToByte(msg);
                        //send the message to the port
                        comPort.Write(newMsg, 0, newMsg.Length);
                        //convert back to hex and display                        
                        DisplayData(MessageType.Outgoing, ByteToHex(newMsg) + ":" + DateTime.Now + "\n");
                    }
                    catch (FormatException ex)
                    {
                        //display error message
                        DisplayData(MessageType.Error, ex.Message);
                    }
                    break;

                case TransmissionType.Text:
                    //first make sure the port is open
                    //if its not open then open it
                    if (!(comPort.IsOpen == true)) comPort.Open();
                    //send the message to the port
                    comPort.Write(msg);
                    //display the message
                    DisplayData(MessageType.Outgoing, msg + DateTime.Now + "\n");
                    break;                    

                case TransmissionType.Hex:
                    try
                    {
                        //convert the message to byte array
                        byte[] newMsg = HexToByte(msg);
                        //send the message to the port
                        comPort.Write(newMsg, 0, newMsg.Length);
                        //convert back to hex and display
                        DisplayData(MessageType.Outgoing, ByteToHex(newMsg) + ":" + DateTime.Now + "\n");
                    }
                    catch (FormatException ex)
                    {
                        //display error message
                        DisplayData(MessageType.Error, ex.Message);
                    }
                    finally
                    {
                       // _displayWindow.SelectAll();
                    }
                    break;
                default:
                    //first make sure the port is open
                    //if its not open then open it
                    if (!(comPort.IsOpen == true)) comPort.Open();
                    //send the message to the port
                    comPort.Write(msg);
                    //display the message
                    DisplayData(MessageType.Outgoing, msg + "\n");
                    break;                    
            }
        }
        #endregion

        #region HexToByte
        /// <summary>
        /// method to convert hex string into a byte array
        /// </summary>
        /// <param name="msg">string to convert</param>
        /// <returns>a byte array</returns>
        private byte[] HexToByte(string msg)
        {
            //remove any spaces from the string
            msg = msg.Replace(" ", "");
            //create a byte array the length of the
            //divided by 2 (Hex is 2 characters in length)
            byte[] comBuffer = new byte[msg.Length / 2];
            //loop through the length of the provided string
            for (int i = 0; i < msg.Length; i += 2)
                //convert each set of 2 characters to a byte
                //and add to the array
                comBuffer[i / 2] = (byte)Convert.ToByte(msg.Substring(i, 2), 16);
            //return the array
            return comBuffer;
        }
        #endregion

        #region ByteToHex
        /// <summary>
        /// method to convert a byte array into a hex string
        /// </summary>
        /// <param name="comByte">byte array to convert</param>
        /// <returns>a hex string</returns>
        private string ByteToHex(byte[] comByte)
        {
            //create a new StringBuilder object
            StringBuilder builder = new StringBuilder(comByte.Length * 3);
            //loop through each byte in the array
            foreach (byte data in comByte)
                //convert the byte to a string and add to the stringbuilder
                builder.Append(Convert.ToString(data, 16).PadLeft(2, '0').PadRight(3, ' '));
            //return the converted value
            return builder.ToString().ToUpper();
        }
        #endregion

        #region DisplayData
        /// <summary>
        /// method to display the data to & from the port
        /// on the screen
        /// </summary>
        /// <param name="type">MessageType of the message</param>
        /// <param name="msg">Message to display</param>
        [STAThread]
        private void DisplayData(MessageType type, string msg)
        {
            _displayWindow.Invoke(new EventHandler(delegate
        {
            _displayWindow.SelectedText = string.Empty;
            _displayWindow.SelectionFont = new Font(_displayWindow.SelectionFont, FontStyle.Bold);
            _displayWindow.SelectionColor = MessageColor[(int)type];
            _displayWindow.AppendText(msg);
            _displayWindow.ScrollToCaret();
        }));
        }
        #endregion

        #region SetActualPosition

        [STAThread]
        private void SetActualPosition(string msg)
        {
            _actualPosition.Invoke(new EventHandler(delegate
            {
               // _actualPosition.SelectedText = string.Empty;
               // _actualPosition.Font = new Font(_actualPosition.Font, FontStyle.Bold);
                _actualPosition.Text = msg;
            }));
        }

        #endregion

        #region OpenPort
        public bool OpenPort()
        {            
            try
            {
                //first check if the port is already open
                //if its open then close it
                if (comPort.IsOpen == true) comPort.Close();

                //set the properties of our SerialPort Object
                comPort.BaudRate = int.Parse(_baudRate);    //BaudRate
                comPort.DataBits = int.Parse(_dataBits);    //DataBits
                comPort.StopBits = (StopBits)Enum.Parse(typeof(StopBits), _stopBits);    //StopBits
                comPort.Parity = (Parity)Enum.Parse(typeof(Parity), _parity);    //Parity
                comPort.PortName = _portName;   //PortName                
                comPort.ReceivedBytesThreshold = _receivedBytes;
                //now open the port
                comPort.Open();
                //display message
                DisplayData(MessageType.Normal, comPort.PortName + " opened at " + DateTime.Now + "\n");               
            
                //return true
                return true;
            }
            catch (Exception ex)
            {
                DisplayData(MessageType.Error, ex.Message + "\n");
                return false;
            }
        }
        #endregion

        #region ClosePort
        public bool ClosePort()
        {
            try
            {
                //first check if the port is already open
                //if its open then close it
                if (comPort.IsOpen == true) comPort.Close();
                //display message
                DisplayData(MessageType.Normal, comPort.PortName + " closed at " + DateTime.Now + "\n");
                //return true if port is closed
                if (comPort.IsOpen == false) return true;
                DisplayData(MessageType.Normal, "Issue closing port\n");
                return false;
            }
            catch (Exception ex)
            {
                DisplayData(MessageType.Error, ex.Message);
                return false;
            }
        }
        #endregion

        #region SetParityValues
        public void SetParityValues(object obj)
        {
            foreach (string str in Enum.GetNames(typeof(Parity)))
            {
                ((ComboBox)obj).Items.Add(str);
            }
        }
        #endregion

        #region SetStopBitValues
        public void SetStopBitValues(object obj)
        {
            foreach (string str in Enum.GetNames(typeof(StopBits)))
            {
                ((ComboBox)obj).Items.Add(str);
            }
        }
        #endregion

        #region SetPortNameValues
        public void SetPortNameValues(object obj)
        {
            // clear combobox 
            ((ComboBox)obj).Items.Clear();
            //write only valid ports into combobox
            foreach (string str in SerialPort.GetPortNames())                        
                ((ComboBox)obj).Items.Add(str);                
        }
        #endregion

        #region comPort_DataReceived
        /// <summary>
        /// method that will be called when theres data waiting in the buffer
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void comPort_DataReceived(object sender, SerialDataReceivedEventArgs e)
        {
            //determine the mode the user selected (binary/string)
            switch (CurrentTransmissionType)
            {
                // only for Haptic Terminal
                case TransmissionType.HT:
                    /*
                    int i = comPort.BytesToRead;                    

                    //read and store data
                    comPort.Read(comBuffer, comBufferEnd, i); 
                    comBufferEnd += i - 1;                                        
                                       
                    while (comBufferStart < comBufferEnd)
                    {                                            
                        // comBuffer enough data for protocol
                        if ((comBufferStart + (FrameLength-1)) <= comBufferEnd)
                        {
                            // valid address at comBuffer[comBufferStart]
                            if (comBuffer[comBufferStart] == 2)
                            {                                        
                                byte CRC = 0;
                                // check CRC                   
                                for (int j = comBufferStart; j < comBufferStart + (FrameLength-1); j++)
                                    CRC += comBuffer[j];

                                if (CRC == comBuffer[comBufferStart + (FrameLength - 1)])
                                {
                                    switch (comBuffer[comBufferStart + 1])
                                    {
                                        case 5:     // get state of the haptic terminal
                                            if ((comBuffer[comBufferStart + 2] == 128) && (comBuffer[comBufferStart + 7] == 1))
                                            {
                                                // send state message to bus
                                                comPort.Write(State, 0, State.Length);                                                    
                                            }
                                            break;
                                        case 6:     // get position data from TMCM and calculate response
                                            if (comBuffer[comBufferStart + 2] == 1)
                                            {                                                                                                
                                                answer[0] = 1;
                                                answer[1] = 2;
                                                answer[2] = 100;
                                                answer[3] = 6;
                                                
                                                // change bitorder to lsb first and convert to int value
                                                Array.Reverse(comBuffer, comBufferStart + 4, 4);
                                                byte_to_int = BitConverter.ToInt32(comBuffer, comBufferStart + 4);
                                                
                                                // set actual position in Form1
                                                SetActualPosition(byte_to_int.ToString());

                                                // calculate left and right limit 
                                                int right_limit = Convert.ToInt32(_rightLimit.Text) - byte_to_int;
                                                int left_limit = Convert.ToInt32(_leftLimit.Text) - byte_to_int;

                                                if (left_limit >= 0)
                                                {
                                                    answer[3] = 6;
                                                }
                                                else if (right_limit >= 0)
                                                {
                                                    answer[3] = 6;
                                                }

                                                // derzeit werden nur die Daten zurck gegeben, es muss noch ein Loader
                                                // fr die Daten geschrieben werden. Erstmals einfach grenzen definieren
                                                answer[4] = comBuffer[comBufferStart + 4];
                                                answer[5] = comBuffer[comBufferStart + 5];
                                                answer[6] = comBuffer[comBufferStart + 6];
                                                answer[7] = comBuffer[comBufferStart + 7];

                                                //calculate CRC
                                                answer[8] = 0;
                                                for (int j=0; j < 9; j++)
                                                    answer[8] += answer[j]; 
                                                
                                                //send the message to the port
                                                comPort.Write(answer, 0, answer.Length);
                                                //convert back to hex and display                        
                                                DisplayData(MessageType.Outgoing, ByteToHex(answer) + ":" + DateTime.Now + "\n");                                                    
                                                }
                                            break;
                                        case 136:   // get version number                                                
                                            //send message to bus
                                            comPort.Write(Version, 0, Version.Length);
                                            break;

                                        default: 
                                            break;
                                    }
                                    // set buffercounter to end of protocol
                                    comBufferStart += FrameLength;
                                }
                                else
                                {
                                    comBufferStart++;
                                }
                            }
                            // no valid address in comBuffer[comBufferStart]
                            else
                            {
                                comBufferStart++;
                            }
                        }
                        // not enough data in comBuffer[]
                        else      
                        {
                            DisplayData(MessageType.Error, "No valid string" + "\n");
                            break;                            
                        }
                    }     
                    // clear position in comBuffer
                    if (comBufferStart == comBufferEnd+1)
                    {
                        comBufferStart = 0;
                        comBufferEnd = 0;
                    }                    
                    break;  */


                    int i = comPort.BytesToRead;
                    
                    //read and store data
                    comPort.Read(comBuffer, 0, i);                    

                    // valid address at comBuffer[comBufferStart]
                    if (comBuffer[0] == 2)
                    {
                        byte CRC = 0;
                        // check CRC                   
                        for (int j = 0; j < FrameLength - 1; j++)
                            CRC += comBuffer[j];

                        if (CRC == comBuffer[FrameLength - 1])
                        {
                            switch (comBuffer[1])
                            {
                                case 5:     // get state of the haptic terminal
                                    if ((comBuffer[2] == 128) && (comBuffer[7] == 1))
                                    {
                                        // send state message to bus
                                        comPort.Write(State, 0, State.Length);
                                    }
                                    break;
                                case 6:     // get position data from TMCM and calculate response
                                    if (comBuffer[2] == 1)
                                    {
                                        answer[0] = 1;
                                        answer[1] = 2;
                                        answer[2] = 100;
                                        answer[3] = 6;

                                        // change bitorder to lsb first and convert to int value
                                        Array.Reverse(comBuffer, 4, 4);
                                        byte_to_int = BitConverter.ToInt32(comBuffer, 4);

                                        // set actual position in Form1
                                        SetActualPosition(byte_to_int.ToString());

                                        // calculate left and right limit 
                                        int right_limit = Convert.ToInt32(_rightLimit.Text) - byte_to_int;
                                        int left_limit = Convert.ToInt32(_leftLimit.Text) - byte_to_int;

                                        if (left_limit >= 0)
                                        {
                                            answer[3] = 6;
                                        }
                                        else if (right_limit >= 0)
                                        {
                                            answer[3] = 6;
                                        }

                                        // derzeit werden nur die Daten zurck gegeben, es muss noch ein Loader
                                        // fr die Daten geschrieben werden. Erstmals einfach grenzen definieren
                                        answer[4] = comBuffer[4];
                                        answer[5] = comBuffer[5];
                                        answer[6] = comBuffer[6];
                                        answer[7] = comBuffer[7];

                                        //calculate CRC
                                        answer[8] = 0;
                                        for (int j = 0; j < 9; j++)
                                            answer[8] += answer[j];

                                        //send the message to the port
                                        comPort.Write(answer, 0, answer.Length);
                                        //convert back to hex and display                        
                                        DisplayData(MessageType.Outgoing, ByteToHex(answer) + ":" + DateTime.Now + "\n");
                                    }
                                    break;
                                case 136:   // get version number                                                
                                    //send message to bus
                                    comPort.Write(Version, 0, Version.Length);
                                    break;

                                default:
                                    break;
                            }                            
                        }
                        else
                        {
                            DisplayData(MessageType.Error, "CRC ERROR" + "\n");
                        }                                      
                    }
                    else 
                    {
                        DisplayData(MessageType.Error, "No valid string" + "\n");
                    }                                      
                    break;

                //user chose string
                case TransmissionType.Text:
                    //read data waiting in the buffer
                    string msg = comPort.ReadExisting();
                    //display the data to the user
                    DisplayData(MessageType.Incoming, msg);
                    break;

                //user chose binary
                case TransmissionType.Hex:
                    //user chose binary
                    //retrieve number of bytes in the buffer
                    int bytes = comPort.BytesToRead;
                    //convert the message to byte array
                    byte[] newMsg = new byte[bytes];                    
                    comPort.Read(newMsg, 0, bytes);
                    //display the data to the user
                    DisplayData(MessageType.Incoming, ByteToHex(newMsg) + ":" + DateTime.Now + "\n");                    
                    break;                    
                default:
                    //read data waiting in the buffer
                    string str = comPort.ReadExisting();
                    //display the data to the user
                    DisplayData(MessageType.Incoming, str + "\n");
                    break;
            }
        }
        #endregion
    }
}
